package Zaphod::Editor;
use strict;
use Carp qw(croak);
use base 'Class::Accessor',
         ;
use List::Util qw(max reduce first);

use Zaphod::Editor::Cursor;

# These plugins are likely unneeded and will
# go in the next release
use Module::Pluggable
    sub_name => 'action_plugins',
    search_path => 'Zaphod::Editor::Action',
    require => 1,
    ;

use Memoize;
memoize 'vocabulary';
use Data::Dumper;

use vars '$VERSION';
$VERSION = '0.02';

=head1 NAME

Zaphod::Editor - document framework for an editor

=cut

__PACKAGE__->mk_accessors(qw(
    undo_list undo_position 
    document cursor 
    log name 
));

#newline
# Do nicer newline support later
# Only on load/save, not while editing!

sub new {
    my ($class, %args) = @_;

    $args{ document } ||= '';
    $args{ undo_list } ||= [];
    $args{ undo_position } ||= 0;
    $args{ 'log' } ||= sub {};
    $args{ name } ||= "unnamed";
    #$args{ newline } ||= "\n";
    my $self = bless \%args, $class;
    $self->{ cursor } ||= Zaphod::Editor::Cursor->new(editor => $self);

    $self
};

=head2 C<< $e->install_verb $verb => $class >>

Registers a verb and its handling class by creating
a subroutine with the name in C<$verb> which
delegates to the class C<$class>. If a subroutine
of that name already exists, the method dies.

=cut

sub install_verb {
    my ($self_or_class,$verb,$delegate) = @_;
    my $class = ref $self_or_class || $self_or_class;
    my $subname = "$class\::$verb";
    no strict 'refs';
    if (defined *{$subname}{CODE}) {
        croak "Won't install $subname in $class - a method of that name already exists";
    };

    *{$subname} = sub {
        my ($self) = shift;
        my $action = $delegate->$verb( $self, @_ );
        $self->record_action($action);
        $action->redo($self);
        $action
    };
};

=head2 C<< __PACKAGE__->install_actions >>

Installs all verbs from all plugins into the
editor class. Inheritance is not really
handled.

=cut

sub install_actions {
    my ($class) = @_;
    my @plugins = $class->action_plugins;

    for my $plugin (@plugins) {
        my @verbs = @{ $plugin->verbs() };
        for my $verb (@verbs) {
            __PACKAGE__->install_verb( $verb => $plugin => $verb );
        };
    };
};

__PACKAGE__->install_actions;

=head2 C<< $e->current_line >>

Convenience C<:lvalue> method to get the
current line

=cut

sub current_line :lvalue {
    $_[0]->document->[ $_[0]->status->line ]
};

=head2 C<< $e->docref >>

Convenience method to get a reference
to the current document. This should be
moved to Data::Alias maybe, to prevent
even further copying of data.

=cut

sub docref { \($_[0]->{document}) }

=head2 C<< $e->record_action ACTION >>

Puts an action onto the undo-stack so it can be
undone later

=cut

sub record_action {
    my ($self,$action) = @_;

    push @{ $self->undo_list }, $action;

    $self->undo_position( scalar @{ $self->undo_list } );
};

=head2 C<< $e->get_actions >>

Returns a copy of all undoable actions

=cut

sub get_actions {
    my ($self) = @_;
    map { $_->copy } @{ $self->undo_list }
};

=head2 C<< $e->apply_actions ACTIONS >>

Apply a list of undoable actions

=cut

sub apply_actions {
    my ($self,@actions) = @_;
    for my $action (@actions) {
        $self->record_action($action);
        $action->redo($self);
    };

    1
};

=head2 C<< $e->undo COUNT >>

Undoes C<COUNT> items in the undo/redo
tree.

=head3 TODO

Change the undo I<list> into a I<tree> so
nested calls of the editor unto itself
undo intuitively.

When passed an array reference, should undo
exactly these actions.

=cut

sub undo {
    my ($self,$count) = @_;
    $count = 1
        unless defined $count;

    my @result;
    # Check that we can actually undo ...
    while ($count--) {
        my $pos = $self->undo_position-1;

        my $action = $self->undo_list->[ $pos ];
        warn "($pos) Undoing " . $action->as_string;
        $action->undo();
        push @result, $action;
        $self->undo_position( $pos );
    };

    @result
};

=head2 C<< $e->redo COUNT >>

Redoes C<COUNT> items in the undo/redo
tree.

When passed an array reference, should redo
exactly these actions.

=cut

sub redo {
    my ($self,$count) = @_;
    $count ||= 1;

    # Check that we can actually undo ...
    my $pos = $self->undo_position+1;

    my $action = $self->undo_list->[ $pos ];
    $action->redo();

    $self->undo_position( $pos );

    $action
};

=head2 C<< $e->can_undo >>

Returns true if the C<undo> action is available,
that is, there are actions that can be undone.

=cut

sub can_undo { $_[0]->undo_position > 0 };

=head2 C<< $e->can_redo >>

Returns true if the C<redo> action is available,
that is, there are actions that can be redone.

=cut

sub can_redo { $_[0]->undo_position < scalar @{ $_[0]->undo_list } };

=head2 C<< $e->is_changed >>

Returns true if the document was changed
since it was last saved.

Currently, this is identical to the document
having undoable actions.

=cut

sub is_changed { $_[0]->undo_position > 0 };

=head2 C<< $e->status >>

Returns an opaque reference encapsulating
the current status of the editor. This is
used by the undo facility to record where
an action takes place.

Currently, this just means the cursor position.

=cut

sub status {
  $_[0]->cursor->copy;
}

=head2 C<< $e->restore_status STATUS >>

Restores the status from a previously saved status.

=cut

sub restore_status {
    my ($self,$status) = @_;
    $self->cursor($status);
};

=head2 C<< $e->vocabulary >>

Returns the list of commands understood
and undoable
by the editor. Mostly convenient for
introspection, but it could be convenient
to use for scripting as well.

=cut

sub vocabulary {
    # For some weird reason there are unnamed classes
    # in the tree. Maybe they come from Memoize
    map { @{ $_->verbs } } grep { !/::::/ } $_[0]->action_plugins;
};

=head2 C<< $e->insert_actions FROM,TO,ACTIONS >>

Replaces the actions in the undo list at
the offsets C<FROM .. TO> with the
actions in C<ACTIONS> and performs
the necessary changes to the document.

=cut

sub insert_actions {
    my ($self,$from,$to,$reconciled) = @_;
    #warn Dumper $reconciled;
    
    my @obsolete = @{ $self->undo_list };
    my @local    = $to < @obsolete ? splice @obsolete, $to+1 : ();
    my @accepted = splice @obsolete, 0, $from -1;
    
    for (reverse @local, reverse @obsolete) {
        $_->revert($self);
    };

    for (@$reconciled) {
        $_->apply($self);
    };

    $self->undo_list([@accepted,@$reconciled]);
    my $u = $self->undo_list;

    my $map = $self->adjustment_map($reconciled);
    for ($self->adjust_change($map,\@local)) {
        $_->apply($self);
        push @$u, $_;
    }
    
    return $map;
}

=head2 C<< $session->adjustment_map CHANGES >>

Creates the list offset changes that result
from applying the given actions.

=cut

sub adjustment_map {
    my ($self,$actions) = @_;
    my %result;

    for my $a (@$actions) {
        my ($ofs,$o,$n) = $a->args;
        my $diff = length($n) - length($o);
        $ofs += length $o;
        if ($diff) {
            $result{$ofs} ||= 0;
            for my $affected (grep { $_ >= $ofs } sort keys %result) {
                $result{ $affected } = $result{ $affected } + $diff;
            };
        };
    };
    \%result;
};

=head2 C<< $e->adjust_change SHIFT_MAP, CHANGES >>

Adjusts the changes by adjusting their offsets
according to the C<SHIFT_MAP> hash reference.

Actions that have no start point anymore are B<not>
yet discarded.

=cut

sub adjust_change {
    my ($self,$map,$changes) = @_;
    #warn Dumper $changes;
    for my $change (@$changes) {
        my ($ofs,$old,$new) = $change->args;
        my $nearest_ofs = (max grep { $_ <= $ofs } keys %$map);
        if (defined $nearest_ofs) {
            $ofs += $map->{$nearest_ofs};
        };
        $change->offset($ofs);
    };
    @$changes;
};

=head2 C<< $e->adjust_offset SHIFT_MAP, OFFSETS >>

Adjusts the offsets
according to the C<SHIFT_MAP> hash reference.

Offsets that don't exist anymore get moved to their
leftmost leftover position.

=cut

sub adjust_offset {
    my ($self,$map,@ofs) = @_;
    #warn Dumper $changes;
    for my $ofs (@ofs) {
        my $nearest_ofs = (max grep { $_ <= $ofs } keys %$map);
        if (defined $nearest_ofs) {
            $ofs += $map->{$nearest_ofs};
        };
    };
    wantarray 
    ? @ofs
    : $ofs[0];
};

=head2 C<< $e->dump >>

Outputs the state of the editor to STDOUT

=cut

sub dump {
    my ($self) = @_;
    my $line = 0;
    print $self->name,"\n";
    printf "%02g|%20s\n", 
        ($line++,$_) for (@{ $self->document });
}

=head2 C<< $e->line_indices >>

Returns a list of all line indices

This is equivalent to

    0..$#{ $e->document }

=cut

sub line_indices {
    0..$#{ $_[0]->document }
};

1;

__END__

=head1 PLUGINS

All plugins for adding new verbs to the editor
are loaded from the Zaphod::Editor::Action:: namespace
and are made available directly in the editor as
undoable actions.

=head1 TODO

=over 4

=item *

Use the logging callback

=item *

Move the installation of verbs from
a (unloggable) class method to
an object method. Or something. So
we get logging.

=item *

Implement de/serialization of undo lists

=item *

Implement hierarchical undo lists so
that we can separate the actions by client.

=back

=cut

=head1 AUTHOR

Written by Max Maischein E<lt>corion@cpan.orgE<gt>

Copyright Max Maischein 2005,2006; All Rights Reserved.

This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zaphod>

=cut
