package Zaphod::Editor::Action;
use strict;
use base qw'Class::Accessor Class::Data::Inheritable';
use Carp qw(croak);
use Scalar::Util qw(weaken);
use Data::Dumper;

=head1 NAME

Zaphod::Editor::Action - encapsulate a change in the editor

=head1 CLASSDATA

=head2 C<< verbs >>

Lists all the verbs that this package exports. Every
verb should be a constructor that returns an instance
of this class which performs and undoes the action.

No constructor should modify the document. The editor
will call the C<apply> method on the action to change
the document.

=cut

use vars qw($VERSION);
$VERSION = '0.02';

__PACKAGE__->mk_classdata('verbs');
__PACKAGE__->verbs([]);
__PACKAGE__->mk_accessors(qw(
    offset
    text_before text_after 
    status_before status_after
));

sub new {
    my ($class,%args) = @_;
    weaken $args{editor};
    bless \%args, $class
};

=head2 C<< $action->apply EDITOR >>

=cut

sub apply {
    my ($self,$editor) = @_;
    my $r = $editor->docref;
    substr( $$r, $self->offset, length $self->text_before ) = $self->text_after;
};

=head2 C<< $action->revert EDITOR >>

=cut

sub revert {
    my ($self,$editor) = @_;
    my $r = $editor->docref;
    substr( $$r, $self->offset, length $self->text_after )= $self->text_before;
};

=head2 C<< $action->copy >>

Returns a (semi-shallow) copy of the action

Copied entities are:

  * offset
  * text_before
  * text_after

=cut

sub copy {
    my ($self) = @_;
    my $copy = bless { %{$self} }, ref $self;
    $copy->{$_} = 
        defined $copy->{$_} 
        ? $copy->{$_}->copy
        : undef
        for qw(status_before status_after);
    $copy
}

=head2 C<< $action->as_js_string >>

Returns the action as a Javascript serialized string.
Currently this doesn't handle escaping properly
- the strings are simply passed through 
quotemeta, which might run afoul of newlines
or HTML chars.

The cursor status of the action is not carried
over, as this is uninteresting.

=cut

sub as_js_string {
    my ($self) = @_;
    sprintf '[%s,%s,%s]', $self->offset, map {$self->encode_js_string($_)} ($self->text_before, $self->text_after);
}

=head2 C<< $action->args >>

A convenience method that returns the
three vital elements of the action. Equivalent
to the following code:

    local $_ = $action;
    return
        $_->offset,
        $_->text_before,
        $_->text_after
    ;

=cut

sub args {
    return
        $_[0]->offset,
        $_[0]->text_before,
        $_[0]->text_after
    ;
}

=head2 C<< Zaphod::Editor::Action->from_string >>

Decodes a list of actions from a string

=cut

=head2 C<< %unquote >>

Mapping of quoted chars to their unquoted internal representation.

Currently maps:

  \r \t \n \" \' \0 \\

=cut

use vars qw(%unquote);

%unquote = map {; $_ => eval qq{"\\$_"} } (qw(r n t " ' 0 \\ ));
for (0..255) {
    $unquote{ sprintf "x%02x", $_ } = chr $_;
}

=head2 C<< unquote ELEMS >>

Unquotes a string according to C<%unquote>. Modifies its
arguments in-place.

=cut

sub unquote {
    for (@_) { 
        s/\\([^x]|x[\dA-Fa-f]{2})/$unquote{$1}/ge; 
    };
};

my $number = qr/(\d+)/;
my $string_dquote = qr/"((?:[^"]*|\\.)*)"/;
my $expr = qr/\[$number,$string_dquote,$string_dquote\]/;

=head2 C<< decode STR >>

Decodes all actions from C<STR>.
If an action is syntactically invalid it is skipped.

=cut

sub decode {
    my $s = $_[0];
    my @result;
    my @values = ($s =~ /(?:^|\G,)$expr/ogc);
    while (@values >= 3) {
        $values[0] = $values[0]+0;
        unquote(@values[1,2]);
        push @result, [ splice @values, 0, 3 ];
    };
    #warn Dumper \@result;
    @result
};

sub from_string {
    my ($class,$editor,$str) = @_;
    #$str =~ s/\$/\\\$/g;
    return map { $class->from_array_ref( $editor, $_ )} decode($str);
}

sub from_array_ref {
    my ($class,$editor,$ar) = @_;
    $class->new(
        offset => $ar->[0],
        text_before => $ar->[1],
        text_after => $ar->[2],
        editor => $editor,
    );
}

sub encode_js_string {
    my $self = shift;
    map {
        s/([\x00-\x1f\\"])/sprintf "\\x%02x", ord $1/gme;
        qq{"$_"}
    } @_
}
1;

=head1 AUTHOR

Written by Max Maischein E<lt>corion@cpan.orgE<gt>

Copyright Max Maischein 2005,2006; All Rights Reserved.

This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zaphod>

=cut
