package Zaphod::Session;
use strict;
use Zaphod::Editor;
use Digest::MD5 qw(md5_base64);
use Data::Dumper;
use Module::Pluggable;
use Carp qw(croak);
use Scalar::Util qw(weaken);
use File::Spec;

use base 'Class::Accessor';

use vars '$VERSION';
$VERSION = '0.02';

require Zaphod::Session::API::JS;
#use Module::Pluggable
#    sub_name => 'available_apis',
#    search_path => 'Zaphod::Session::API',
#    require => 1
#    ;

__PACKAGE__->mk_accessors(qw( 
    clients
    editor
    state
    cookie
    secret
    apis
    timeout
    name
));

=head2 C<< Zaphod::Session->new ARGS >>

Creates a new session.

The following fields can be set through the
constructor:

    clients  - hashref of preexisting clients
    document - arrayref of the document
    timeout  - client session timeout in seconds
    cookie   - the secret cookie
    name     - the session name

=cut

sub new {
    my ($package, %args) = @_;

    $args{ document } ||= '';
    if (ref $args{document}) {
        open my $fh, "<:utf8", $args{document}
            or die "Couldn't open '$args{document}': $!";
        local $/;
        $args{document} = <$fh>;
    };
    $args{ editor }   ||= Zaphod::Editor->new(
        document => delete $args{ document },
    );
    $args{ clients }    ||= {};
    $args{ state }      ||= 0;
    $args{ cookie }     ||= 'The One Session';
    $args{ secret }     ||= 'Zaphod Secret Sauce';
    $args{ timeout }    ||= 5;
    # when, what, who
    $args{ name }       ||= [ $args{state}, "Unnamed", undef ];
    
    #$args{ send_to_client } = sub {};

    my $self = $package->SUPER::new(\%args);
    
    my $w = $self;
    weaken $w;
    my %apis = map { 
            $_ =~ /::(\w+)$/
                or croak "Weird plugin name: " . $_;
            $1 => $_->new({session => $w});
        } (qw[Zaphod::Session::API::JS]);
    $self->apis(\%apis);
    #warn Dumper $args{apis};

    $self;
}

=head2 C<< $session->log TYPE, MSG >>

Logging callback. Currently just prints to C<STDOUT>

=cut

sub log {
    my ($self,$type,@messages) = @_;
    for (@messages) {
        s/^/$type: /gm;
        warn $_;
    }
}

=head2 C<< $session->api API >>

Factory method to produce and instantiate the APIs. Currently,
the only supported API is

=cut

sub api {
    my ($self,$api) = @_;
    $self->apis->{$api}
}

=head2 C<< $session->action_from_client CLIENT,STATE,ACTIONS >>

Insert actions from a client, advancing the
master state.

Returns a true value if all actions applied without problems.
If there were some actions that couldn't be merged without
a conflict, returns a false value. All other actions got
applied.

The conflicting action is silently discarded at the moment,
at least there's no feedback mechanism for the client yet.

=cut

sub action_from_client {
    my ($self,$client,$state,@actions) = @_;

    # What would I really need the client for?
    
    # 1. Transform the client's actions from the state they were made
    # in to the current state

    # 2. Add the client's actions:
    # Prior to adding, these should be modified by all
    # actions received since $state instead.
    # Then the actions can simply be appended.

    my $e = $self->editor;
    my $changes = $e->undo_list;
    my @other_changes = @{$changes}[$state+1..$#$changes];
    my $map = $e->adjustment_map(\@other_changes);
    for ($e->adjust_change($map,\@actions)) {
        # Need a check if the change can be applied at all
        $_->apply($e);
        $e->record_action($_);
        warn "Editor state now " . scalar @{ $e->undo_list };
        #push @resulting_changes, $_;
    }
    #my $map = $self->editor->insert_actions( $state, $state, \@actions );

    # 3. Update all client positions by these actions, save the
    # current client. These will of course be rectified by the
    # other clients as soon as they poll. In theory, these updates
    # could be handled completely locally when applying the changes.

    my $cl = $self->clients;
    for my $c ($self->users_for_client($client)) {
        my $old = $cl->{$c}->{cursor_position};
        my $new = $self->editor->adjust_offset($map,$old);
        $cl->{$c}->{cursor_position} = $new;
    }

    # Everything between $state and $self->state()
    # is what needs to be sent back

    $self->state( $e->undo_position );
    $self->saw_client( $client );
    
    $self->state;
}

=head2 C<< $session->actions_for_client CLIENT, STATE >>

Returns the actions that bring the client from its
current state to the current state of the master.

=cut

sub actions_for_client {
    my ($self,$client,$state) = @_;
    
    # The current state is, "obviously", the number of actions
    # inserted
    
    $state ||= $self->clients->{$client}->{state} || 0;

    my @a = $self->editor->get_actions;
    
    warn "Updating client from $state to " . scalar @a;
    
    if ($state < $self->state) {
        return splice @a, $state;
    } else {
        return ()
    }
}

=head2 C<< $session->gen_cookie CLIENT >>

Generates a shared secret for identifying
a connected client.

=cut

sub gen_cookie {
    my ($self,$visual) = @_;
    
    # This should become configurable, later,
    # or simply initialized at server startup
    # to some random string
    my $secret = 'Zaphod Secret Sauce';
    my $cookie = $self->cookie;

    return md5_base64( "$secret$cookie$$" . rand(0x10000));
}

sub gen_username {
    my ($self) = @_;
    my $templ = "Unnamed %s";
    my %n = map { $_->{visual} => 1 } values %{ $self->clients };
    my $count = 0;
    my $name;
    
    do {
        $count++;
        $name = sprintf $templ, $count;
    } until not exists $n{ $name };
    warn "Generated $name";
    return $name;
}

=head2 C<< $session->gen_color >>

Returns a free color number

=cut

sub gen_color {
    my ($self) = @_;
    my %c = map { $_->{color} => 1 } values %{ $self->clients };
    my $color = 0;
    
    do {
        $color++;
    } until not exists $c{ $color };
    return $color;
}

=head2 C<< $session->valid_cookie COOKIE >>

Returns true if the given cookie is a valid
client connection. In fact, the value returned
is the hash entry for the relevant client.

=cut

sub valid_cookie { $_[0]->clients->{$_[1]} };

=head2 C<< $session->connect_client VISUAL >>

Adds/connects a new client.

The current client attributes are

  state           - state of the client document
  last_seen       - epoch when the client last connected
  cursor_position - offset of the client position
  visual          - the name of the client

The function returns the handle to be used in
all future function calls of the client.

=cut

sub connect_client {
    my ($self,$visual) = @_;
    
    # First, sanitize the visual
    $visual = $self->sane_username($visual);
    
    while (! $self->unique_username($visual)) {
        $self->log("STATUS",sprintf "Name clash. Changing %s to %s connected", $visual, $visual . "_");
        $visual .= "_";
    };
    
    my $client = $self->gen_cookie($visual);
    my $color = $self->gen_color;
    $self->clients->{$client} = {
        visual          => $visual,
        state           => undef,
        last_seen       => undef,
        cursor_position => 0,
        color           => $color,
    };
    $self->saw_client( $client );
    $self->client_cursor( $client, 0 );
    $self->log("STATUS",sprintf "%s connected", $visual);
    return $client
}
    
=head2 C<< $session->sane_username VISUAL >>

Returns true if the username C<VISUAL>
is considered sane enough.

Currently this means it doesn't start with, end with
or contain multiple whitespace characters. 

HTML entities are not yet handled (see TODO)

=cut
    
sub sane_username {
    my ($self,$visual) = @_;
    $visual =~ s/\s+/ /g;
    $visual =~ s/\s*$//;
    $visual =~ s/^\s*//;
    $visual;
}

=head2 C<< $session->unique_username VISUAL >>

Returns true if the username C<VISUAL> is considered
unique enough.

=cut

sub unique_username {
    my ($self,$visual) = @_;
    
    my $c = $self->clients;
    # Check that we have a unique visual by doing a case match
    if (grep { lc $_->{visual} eq lc $visual } (values %$c)) {
        return
    }
    return 1
}

=head2 C<< $session->saw_client CLIENT >>

Resets the timeout counter for C<CLIENT>
if it exits. Returns the hashref
corresponding to C<CLIENT> if it exists.

=cut

sub saw_client {
    my ($self,$client) = @_;
    my $c = $self->valid_cookie($client);
    if (defined $c) {
        $c->{last_seen} = time();
    };
    
    $self->prune_clients();
    
    $c
}

=head2 C<< $session->user_count >>

Returns the number of logged in users

=cut

sub user_count {
    my ($self) = @_;
    my $c = $self->clients;
    scalar keys %{ $self->clients }
}

=head2 C<< $session->client_cursor CLIENT, OFFSET >>

Sets the offset for the cursor position of the client

=cut

sub client_cursor {
    my ($self,$client,$offset) = @_;
    
    $offset = 0
        if $offset < 0;

    if (my $c = $self->saw_client($client)) {
        $c->{cursor_position} = $offset;
    };
}

=head2 C<< $session->disconnect_client CLIENT >>

Removes the client from the session

=cut

sub disconnect_client {
    my ($self,$client) = @_;
    delete $self->clients->{$client};
}

=head2 C<< $session->prune_clients TIMEOUT >>

Disconnects all clients that haven't polled
since C<TIMEOUT> seconds. If C<TIMEOUT> is not given,
the value of the C<timeout> accessor is used.

If a disconnected client later connects again,
the connection is rejected and the client
will have to log in again.

=cut

sub prune_clients {
    my ($self,$timeout) = @_;
    $timeout ||= $self->timeout;
    my $cutoff = time - $timeout;
    my $c = $self->clients;
    
    for my $d (grep { $c->{$_}->{last_seen} < $cutoff } keys %$c) {
        $self->log("STATUS",sprintf "Disconnecting %s (timeout after $timeout seconds)", $c->{$d}->{visual});
        $self->disconnect_client($d);
    }
}

=head2 C<< $session->users_for_client CLIENT >>

Returns the list of the other users and their
positions for the session

=cut

sub users_for_client {
    my ($self,$client) = @_;
    my $c = $self->clients;
    grep { $_ ne $client } sort keys %$c;
}

=head2 C<< $session->set_name NEWNAME >>

Sets the name of the current session, permissions provided.

The new name will be broadcast to the other clients.
The name change is currently out-of-band with the editor,
so there is no undo/redo functionality for it.

=cut

sub set_name {
    my ($self,$client,$name,$state) = @_;
    my ($last_state,$last_name) = @{ $self->name };
    if ($self->state >= $state) {
        $self->name( [$self->state, $name, $client] );
    }
};

=head2 C<< $session->save DIR >>

Serializes the session to disk. This will create two
files in DIR, C<$session.document> and C<$session.meta>.
The C<.document> file will contain the document,
and C<.meta> file will contain metadata. The following metadata
is saved:

    state
    cookie
    secret
    name

If more elaborate session management comes to frutition,
also the session owner and allowed viewers etc. will
be saved in the C<.meta> file. The C<.meta> file is
a simple custom file format of key-value pairs
matching

  /^(\w+)=(.*)$/

and will be read and written as UTF-8. The document will
be read and written via C<:utf8> as well.

=cut

sub save {
    my ($self,$dir) = @_;
    my $docname = File::Spec->catfile( $dir, $self->cookie . ".document" );
    my $metaname = File::Spec->catfile( $dir, $self->cookie . ".meta" );
    
    # Save the document
    open my $fh, ">:utf8", $docname
        or die "Couldn't create '$docname': $!";
    binmode $fh;
    print {$fh} $self->editor->document;
    close $fh;
    
    # Save the metadata
    open $fh, ">:utf8", $metaname
        or die "Couldn't create '$metaname': $!";
    binmode $fh;
    
    for my $key ($self->persistent_values) {
        print {$fh} sprintf "%s=%s\n", $key, $self->$key
    };
    print {$fh} "name=" .  $self->name->[1] . "\n";
    close $fh;
}

=head2 C<< Zaphod::Session->load FILE, ARGS >>

Constructor to intialize and then load a session
from a file.

The filename can be either the name of the document
file or the name of the metafile.

Currently, C<persistent_values> is not checked against
the keys loaded from the C<.meta> file.

=cut

sub load {
    my ($class,$file,%args) = @_;
    (my $name = $file) =~ s/(\.document|\.meta)$//i;
    my $docname = "$name.document";
    my $metaname = "$name.meta";
    
    # Load the metadata
    open my $fh, "<:utf8", $metaname
        or die "Couldn't create '$metaname': $!";
    binmode $fh;
    
    while (<$fh>) {
        chomp;
        next if /^#/;
        if (/^name=(.*)$/) {
            warn "Setting name to $1";
            $args{name} = [$args{state},$1,"<restored from session>"];
        } elsif (/^(\w+)=(.*)$/) {
            $args{$1} = $2;
        } else {
            croak "Couldn't load '$metaname': Illegal line format: $_";
        }
    };
    close $fh;

    local $/;
    
    # Load the document
    open $fh, "<:utf8", $docname
        or die "Couldn't open '$docname': $!";
    binmode $fh;    
    $args{document} = <$fh>;
    close $fh;
    
    warn sprintf "Loaded session '%s' from '%s'",
        $args{name}->[1], $metaname;
    
    $class->new(%args);
}

=head2 C<< $session->persistent_values >>

(Class method)

Returns all keys except C<document> that persist on-disk.

=cut

sub persistent_values {qw(
    state
    cookie
    secret
)}

1;

=head1 AUTHOR

Written by Max Maischein E<lt>corion@cpan.orgE<gt>

Copyright Max Maischein 2005,2006; All Rights Reserved.

This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zaphod>

=cut
